import datetime
import logging
from typing import List, Set, Tuple

from httpx import URL


from ollama_downloader.data.data_models import ImageManifest
from ollama_downloader.downloader.model_downloader import ModelDownloader, ModelSource
import lxml.html
from ollama import Client as OllamaClient


from rich import print as print

# Initialize the logger
logger = logging.getLogger(__name__)


class OllamaModelDownloader(ModelDownloader):
    def __init__(self):
        super().__init__()

    def download_model(self, model_identifier: str) -> bool:
        model, tag = (
            model_identifier.split(":")
            if ":" in model_identifier
            else (model_identifier, "latest")
        )
        print(f"Downloading Ollama library model {model}:{tag}")
        # Validate the response as an ImageManifest but don't enforce strict validation
        manifest_json = self._fetch_manifest(
            model_identifier=model_identifier, model_source=ModelSource.OLLAMA
        )
        logger.info(f"Validating manifest for {model}:{tag}")
        manifest = ImageManifest.model_validate_json(manifest_json, strict=True)
        # Keep a list of files to be copied but only copy after all downloads have completed successfully.
        # This is to ensure that we don't copy files that may not be needed if the download fails.
        # Each tuple in the list contains (source_path, named_digest, computed_digest).
        files_to_be_copied: List[Tuple[str, str, str]] = []
        # Download the model configuration BLOB
        logger.info(f"Downloading model configuration {manifest.config.digest}")
        file_model_config, digest_model_config = self._download_model_blob(
            model_identifier=model_identifier,
            named_digest=manifest.config.digest,
            model_source=ModelSource.OLLAMA,
        )
        files_to_be_copied.append(
            (file_model_config, manifest.config.digest, digest_model_config)
        )
        # The layers could be null for cloud-hosted models.
        if manifest.layers:
            for layer in manifest.layers:
                logger.debug(
                    f"Layer: {layer.mediaType}, Size: {layer.size} bytes, Digest: {layer.digest}"
                )
                logger.info(f"Downloading {layer.mediaType} layer {layer.digest}")
                file_layer, digest_layer = self._download_model_blob(
                    model_identifier=model_identifier,
                    named_digest=layer.digest,
                    model_source=ModelSource.OLLAMA,
                )
                files_to_be_copied.append((file_layer, layer.digest, digest_layer))
        # All BLOBs have been downloaded, now copy them to their appropriate destinations.
        for source, named_digest, computed_digest in files_to_be_copied:
            copy_status, copy_destination = self._save_blob(
                source=source,
                named_digest=named_digest,
                computed_digest=computed_digest,
            )
            if copy_status is False:
                raise RuntimeError(
                    f"Failed to copy {named_digest} to {copy_destination}."
                )
        # Finally, save the manifest to its appropriate destination
        self._save_manifest(
            data=manifest_json,
            model_identifier=model_identifier,
            model_source=ModelSource.OLLAMA,
        )
        ts_approximate_manifest_save = datetime.datetime.now()
        # Finally check if it exists in the Ollama
        # Clear the list of unnecessary files before this if errors henceforth are to be tolerated.
        if not self.settings.ollama_server.remove_downloaded_on_error:
            self._unnecessary_files.clear()
        ollama_client = OllamaClient(
            host=self.settings.ollama_server.url,
            # timeout=self.settings.ollama_server.timeout,
            # TODO: Add API key authentication logic
        )
        models_list = ollama_client.list()
        found_model = None
        for model_info in models_list.models:
            if (
                model_info.model == f"{model}:{tag}"
                # TODO: Is this timestamp assumption right that the listing is completed within a minute of saving?
                and abs(
                    model_info.modified_at.replace(tzinfo=None)
                    - ts_approximate_manifest_save
                )
                < datetime.timedelta(minutes=1)
            ):
                found_model = model_info
                break
        if found_model:
            print(
                f"Model {found_model.model} successfully downloaded and saved on {found_model.modified_at:%B %d %Y at %H:%M:%S}."
            )
        else:
            raise RuntimeError(
                f"Model {model}:{tag} could not be found in Ollama server after download."
            )
        # If we reached here cleanly, remove all unnecessary file names but don't remove actual files.
        self._unnecessary_files.clear()
        # Finally return success
        return True

    def list_available_models(
        self, page: int | None = None, page_size: int | None = None
    ) -> List[str]:
        with self.get_httpx_client(
            verify=self.settings.ollama_library.verify_ssl,
            timeout=self.settings.ollama_library.timeout,
        ) as client:
            logger.debug(
                f"Updating models list from Ollama library {self.settings.ollama_library.library_base_url}"
            )
            models_response = client.get(self.settings.ollama_library.library_base_url)
            models_response.raise_for_status()
            parsed_models_html = lxml.html.document_fromstring(models_response.text)
            available_models = []
            library_prefix = "/library/"
            for _, attribute, link, _ in lxml.html.iterlinks(parsed_models_html):
                if attribute == "href" and link.startswith(library_prefix):
                    available_models.append(link.replace(library_prefix, ""))
            logger.debug(f"Found {len(available_models)} models in the Ollama library.")
            available_models.sort(key=lambda s: s.lower())
            paginated_result = available_models
            if page_size and page:
                # Adjust page number for 0-based index
                start_index = (page - 1) * page_size
                end_index = start_index + page_size
                paginated_result = available_models[start_index:end_index]
            if len(paginated_result) == 0:
                logger.warning(
                    f"No models found for the specified page {page} and page size {page_size}. Returning all models instead."
                )
                paginated_result = available_models
                page = None
            return paginated_result

    def list_model_tags(self, model_identifier: str) -> List[str]:
        available_models = self.list_available_models()
        if model_identifier not in available_models:
            raise RuntimeError(
                f"Model {model_identifier} not found in the library models list."
            )
        with self.get_httpx_client(
            verify=self.settings.ollama_library.verify_ssl,
            timeout=self.settings.ollama_library.timeout,
        ) as client:
            models_tags = []
            logger.debug(
                f"Fetching tags for model {model_identifier} from the Ollama library."
            )
            tags_response = client.get(
                URL(self.settings.ollama_library.library_base_url).join(
                    f"{model_identifier}/tags"
                )
            )
            tags_response.raise_for_status()
            logger.debug(f"Parsing tags for model {model_identifier}.")
            parsed_tags_html = lxml.html.document_fromstring(tags_response.text)
            library_prefix = "/library/"
            named_model_unique_tags: Set[str] = set()
            for _, attribute, link, _ in lxml.html.iterlinks(parsed_tags_html):
                if attribute == "href" and link.startswith(
                    f"{library_prefix}{model_identifier}:"
                ):
                    named_model_unique_tags.add(link.replace(library_prefix, ""))
            models_tags = list(named_model_unique_tags)
            models_tags.sort(key=lambda s: s.lower())

            return models_tags



# Work-in-progress script to filter pstats data to only include relevant functions.
import os
import pstats
import re
import subprocess
from rich import print as print


def filter_pstats_and_save_svg(
    input_file: str, output_file: str, keep_regexes: list[str]
):
    """
    Loads pstats data, filters it, and saves a new pstats file.
    `keep_regexes` is a list of regex patterns to match against function names.
    """
    stats = pstats.Stats(input_file)
    # FIXME: Get the raw dictionary of stats -- this is very hacky because pstats doesn't expose this directly.
    original_stats = stats.stats  # type: ignore[attr-defined]

    filtered_stats = {}

    # Compile the regexes for efficiency
    compiled_regexes = [re.compile(r) for r in keep_regexes]
    print(f"Compiled regexes: {len(compiled_regexes)}")

    # Find all functions that match your regexes
    to_keep = set()
    for func, (cc, nc, tt, ct, callers) in original_stats.items():
        # func is a tuple like ('my_package/my_module.py', 10, 'my_function')
        func_name = str(func)
        if any(regex.search(func_name) for regex in compiled_regexes):
            to_keep.add(func)

    # Include all dependencies of the kept functions
    all_needed = set(to_keep)
    queue = list(to_keep)
    print("Matching functions:")
    print(queue)

    while queue:
        func = queue.pop(0)
        (cc, nc, tt, ct, callers) = original_stats.get(func, (0, 0, 0, 0, {}))
        for caller_func in callers.keys():
            if caller_func not in all_needed:
                all_needed.add(caller_func)
                queue.append(caller_func)

    # Build the new stats dictionary, including only the functions we need
    for func in all_needed:
        if func in original_stats:
            filtered_stats[func] = original_stats[func]

    stats.stats = filtered_stats  # type: ignore[attr-defined]
    stats.dump_stats(output_file)


def main():
    # Define the patterns to keep. This should capture your tests and application code.
    patterns_to_keep = [
        # Your specific test files or modules
        "tests/",
        "src/ollama_downloader",
        # Any other modules you want to include
        # "src/other_module"
    ]

    # Filter and save the new profile data
    PROF_DIR = "prof"
    TARGET_PROF = "combined.prof"
    OUTPUT_PROF = "filtered.prof"
    OUTPUT_SVG = "filtered_profile.svg"
    target_path = os.path.join(PROF_DIR, TARGET_PROF)
    try:
        if not os.path.exists(target_path):
            raise FileNotFoundError(
                f"Profile data file {target_path} not found. Maybe run the tests with profiling first?"
            )
        output_path = os.path.join(PROF_DIR, OUTPUT_PROF)
        output_svg_path = os.path.join(PROF_DIR, OUTPUT_SVG)
        filter_pstats_and_save_svg(target_path, output_path, patterns_to_keep)
        print(f"Filtered profile data saved to {output_path}.")

        conversion_command = (
            f"gprof2dot -f pstats {output_path} | dot -Tsvg -o {output_svg_path}"
        )
        result = subprocess.run(
            conversion_command, capture_output=True, shell=True, check=True
        )
        print(f"{result.args} completed with return code {result.returncode}.")
        if result.returncode != 0:
            print(f"stderr: {result.stderr.decode()}")
            print(f"stdout: {result.stdout.decode()}")
    except Exception as e:
        print(f"Error during filtering and conversion of profile data. {e}")


if __name__ == "__main__":
    main()



import datetime
import logging
from typing import Annotated, List, Optional, Tuple
from urllib.parse import urlparse


from pydantic import Field

from ollama_downloader.data.data_models import ImageManifest
from ollama_downloader.downloader.model_downloader import ModelDownloader, ModelSource

# import lxml.html
from ollama import Client as OllamaClient
# from huggingface_hub import configure_http_backend
# import requests  # type: ignore

# Initialize the logger
logger = logging.getLogger(__name__)


class HuggingFaceModelDownloader(ModelDownloader):
    def __init__(self):
        super().__init__()

    def download_model(self, model_identifier: str) -> bool:
        # Validate the response as an ImageManifest but don't enforce strict validation
        (user, model_repo), quant = (
            model_identifier.split(":")[0].split("/"),
            model_identifier.split(":")[1] if ":" in model_identifier else "latest",
        )
        print(
            f"Downloading Hugging Face model {model_repo} from {user} with {quant} quantisation"
        )
        manifest_json = self._fetch_manifest(
            model_identifier=model_identifier, model_source=ModelSource.HUGGINGFACE
        )
        logger.info(f"Validating manifest for {model_identifier}")
        manifest = ImageManifest.model_validate_json(manifest_json, strict=True)
        # Keep a list of files to be copied but only copy after all downloads have completed successfully.
        # This is to ensure that we don't copy files that may not be needed if the download fails.
        # Each tuple in the list contains (source_path, named_digest, computed_digest).
        files_to_be_copied: List[Tuple[str, str, str]] = []
        # Download the model configuration BLOB
        logger.info(f"Downloading model configuration {manifest.config.digest}")
        file_model_config, digest_model_config = self._download_model_blob(
            model_identifier=model_identifier,
            named_digest=manifest.config.digest,
            model_source=ModelSource.HUGGINGFACE,
        )
        files_to_be_copied.append(
            (file_model_config, manifest.config.digest, digest_model_config)
        )
        # The layers could be null for cloud-hosted Ollama models but this is here only for consistency.
        if manifest.layers:
            for layer in manifest.layers:
                logger.debug(
                    f"Layer: {layer.mediaType}, Size: {layer.size} bytes, Digest: {layer.digest}"
                )
                logger.info(f"Downloading {layer.mediaType} layer {layer.digest}")
                file_layer, digest_layer = self._download_model_blob(
                    model_identifier=model_identifier,
                    named_digest=layer.digest,
                    model_source=ModelSource.HUGGINGFACE,
                )
                files_to_be_copied.append((file_layer, layer.digest, digest_layer))
        # All BLOBs have been downloaded, now copy them to their appropriate destinations.
        for source, named_digest, computed_digest in files_to_be_copied:
            copy_status, copy_destination = self._save_blob(
                source=source,
                named_digest=named_digest,
                computed_digest=computed_digest,
            )
            if copy_status is False:
                raise RuntimeError(
                    f"Failed to copy {named_digest} to {copy_destination}."
                )
        # Finally, save the manifest to its appropriate destination
        self._save_manifest(
            data=manifest_json,
            model_identifier=model_identifier,
            model_source=ModelSource.HUGGINGFACE,
        )
        ts_approximate_manifest_save = datetime.datetime.now()
        # Finally check if it exists in the Ollama
        # Clear the list of unnecessary files before this if errors henceforth are to be tolerated.
        if not self.settings.ollama_server.remove_downloaded_on_error:
            self._unnecessary_files.clear()
        ollama_client = OllamaClient(
            host=self.settings.ollama_server.url,
            # timeout=self.settings.ollama_server.timeout,
            # TODO: Add API key authentication logic
        )
        models_list = ollama_client.list()
        found_model = None
        search_model = (
            f"{urlparse(ModelDownloader.HF_BASE_URL).hostname}/{model_identifier}"
        )
        for model_info in models_list.models:
            if (
                model_info.model == search_model
                # TODO: Is this timestamp assumption right that the listing is completed within a minute of saving?
                and abs(
                    model_info.modified_at.replace(tzinfo=None)
                    - ts_approximate_manifest_save
                )
                < datetime.timedelta(minutes=1)
            ):
                found_model = model_info
                break
        if found_model:
            print(
                f"Model {found_model.model} successfully downloaded and saved on {found_model.modified_at:%B %d %Y at %H:%M:%S}."
            )
        else:
            raise RuntimeError(
                f"Model {search_model} could not be found in Ollama server after download."
            )
        # If we reached here cleanly, remove all unnecessary file names but don't remove actual files.
        self._unnecessary_files.clear()
        return True

    def list_available_models(
        self,
        page: Annotated[Optional[int], Field(gt=0)] = None,
        page_size: Annotated[Optional[int], Field(gt=0, le=100)] = None,
    ) -> List[str]:
        page_size = page_size or 100
        next_page = 1
        page = page or 1
        if page_size * (page + 1) >= 1000:
            logger.warning(
                "Hugging Face currently does not allow paging beyond the first 999 models. Follow issue 2741: https://github.com/huggingface/huggingface_hub/issues/2741"
            )
            raise ValueError(
                f"Hugging Face currently does not allow obtaining information beyond the first 999 models. Your requested page {page} with page size {page_size} exceeds this limit by {int(((page + 1) * page_size - 999))} model(s)."
            )
        api_url = f"https://huggingface.co/api/models?apps=ollama&gated=false&limit={page_size}&sort=trendingScore"
        next_page_url = api_url

        with self.get_httpx_client(
            self.settings.ollama_library.verify_ssl,
            self.settings.ollama_library.timeout,
        ) as client:
            while next_page < page and next_page_url:
                models_head = client.head(next_page_url)
                models_head.raise_for_status()
                next_page_url = models_head.links.get("next", {}).get("url")
                next_page += 1
            if next_page_url:
                if next_page > 1:
                    logger.info(f"Requesting page {next_page} from {next_page_url}")
                models_response = client.get(next_page_url)
                models_response.raise_for_status()
            model_identifiers = [
                model["modelId"] for model in list(models_response.json())
            ]
        logger.warning(
            "HuggingFace models are sorted in the context of the selected page only."
        )
        model_identifiers.sort(key=lambda s: s.lower())
        return model_identifiers

    def list_model_tags(self, model_identifier: str) -> List[str]:
        api_url = f"https://huggingface.co/api/models/{model_identifier}?blobs=true"
        tags = []
        with self.get_httpx_client(
            self.settings.ollama_library.verify_ssl,
            self.settings.ollama_library.timeout,
        ) as client:
            model_response = client.get(api_url)
            model_response.raise_for_status()
            model_info = model_response.json()
            model_siblings = model_info.get("siblings", [])
            for repo_sibling in model_siblings:
                rfilename = repo_sibling.get("rfilename", "")
                if rfilename.endswith(".gguf"):
                    # Try to extract the quantisation from the filename
                    tag = rfilename.split(".gguf")[0].split("-")[-1]
                    tags.append(f"{model_identifier}:{tag}")
        if len(tags) == 0:
            # If no .gguf files found, the model is not for Ollama
            raise RuntimeError(
                f"The model {model_identifier} has no support for Ollama."
            )
        tags.sort(key=lambda s: s.lower())
        return tags



import logging
import os
import re
from typing import ClassVar
import platform
import psutil

try:
    import grp  # Unix, macOS and Linux only
except ImportError:
    # Windows does not have the grp module
    grp = None  # type: ignore[assignment]

from ollama import Client as OllamaClient

logger = logging.getLogger(__name__)


class OllamaSystemInfo:
    """Experimental class to obtain system information related to Ollama."""

    PROCESS_NAME: ClassVar[str] = "ollama"
    _instance: ClassVar = None

    os_name: str = ""
    process_id: int = -1
    process_env_vars: dict[str, str] = {}
    parent_process_id: int = -1
    process_owner: tuple[str, int, str, int] = (
        "",
        -1,
        "",
        -1,
    )  # (username, uid, groupname, gid)
    listening_on: str = ""  # e.g. "http://localhost:11434"
    models_dir_path: str = ""
    likely_daemon: bool = False

    def __new__(cls: type["OllamaSystemInfo"]) -> "OllamaSystemInfo":
        """
        The singleton pattern is used to ensure that calls to certain methods are cached
        using instance variables. However, only one instance of this class should exist.
        """
        if cls._instance is None:
            # Create instance using super().__new__ to bypass any recursion
            instance = super().__new__(cls)
            cls._instance = instance
        return cls._instance

    def is_windows(self) -> bool:
        """Check if the operating system is Windows."""
        if self.os_name == "":
            self.os_name = platform.system()
        return self.os_name.lower() == "windows"

    def is_macos(self) -> bool:
        """Check if the operating system is macOS."""
        if self.os_name == "":
            self.os_name = platform.system()
        return self.os_name.lower() == "darwin"

    def is_running(self) -> bool:
        """
        Check if the Ollama process is running on the system.
        This will not work if Ollama is running in a container.
        """
        if self.process_id == -1:
            for proc in psutil.process_iter(["pid", "name"]):
                if (
                    proc.info["name"]
                    and
                    # Use exact match to avoid false positives from substring matches but what about Windows, e.g., "ollama.exe"?
                    OllamaSystemInfo.PROCESS_NAME.lower() == proc.info["name"].lower()
                ):
                    self.process_id = proc.info["pid"]
                    break
            if self.process_id != -1:
                logger.debug(f"Ollama process found with PID {self.process_id}.")
                try:
                    self.process_env_vars = {}
                    proc = psutil.Process(self.process_id)
                    # FIXME: These will not capture any variables that the Ollama process sets after it starts.
                    # For example, "OLLAMA_MODELS" is not captured this way unless explicitly passed.
                    self.process_env_vars.update(proc.environ())
                    if len(self.process_env_vars) > 0:
                        logger.debug(
                            f"{len(self.process_env_vars)} environment variables of process {proc.name()} ({self.process_id}, {proc.status()}) were obtained."
                        )
                except psutil.NoSuchProcess:
                    ...
                except psutil.AccessDenied:
                    logger.warning(
                        f"Environment variables of process {proc.name()} ({self.process_id}, {proc.status()}) cannot be retrieved. Run auto-config as super-user."
                    )
            else:
                logger.warning(
                    "Ollama process not found. Maybe, it is not installed or it is not running."
                )
        return self.process_id != -1

    def get_parent_process_id(self) -> int:
        """
        Get the parent process ID of the Ollama process.
        This will fail if Ollama is running as a service and this function is called by not a super-user.
        """
        if self.parent_process_id == -1 and self.is_running():
            try:
                proc = psutil.Process(self.process_id)
                self.parent_process_id = proc.ppid()
            except psutil.NoSuchProcess:
                ...
            except psutil.AccessDenied:
                logger.warning(
                    f"Parent process ID of process {proc.name()} ({self.process_id}, {proc.status()}) cannot be retrieved. Run auto-config as super-user."
                )
        return self.parent_process_id

    def get_process_owner(self) -> tuple[str, int, str, int] | None:
        """Get the owner of the Ollama process as a tuple of (username, uid, groupname, gid)."""
        if self.process_owner == ("", -1, "", -1) and self.is_running():
            try:
                proc = psutil.Process(self.process_id)
                username = proc.username()
                uid = proc.uids().real if hasattr(proc, "uids") else -1
                gid = proc.gids().real if hasattr(proc, "gids") else -1
                groupname = (
                    grp.getgrgid(gid).gr_name if grp is not None and gid != -1 else ""
                )
                self.process_owner = (username, uid, groupname, gid)
                logger.debug(
                    f"Owner of process {proc.name()} ({self.process_id}, {proc.status()}): {self.process_owner}"
                )
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                ...
        return self.process_owner

    def is_model_dir_env_var_set(self) -> bool:
        """Check if the environment variable 'OLLAMA_MODELS' is set in the Ollama process."""
        if self.process_env_vars:
            return "OLLAMA_MODELS" in self.process_env_vars
        return False

    def infer_listening_on(self) -> str | None:
        """Get the address and port the Ollama process is listening on."""
        if self.listening_on == "" and self.is_running():
            try:
                proc = psutil.Process(self.process_id)
                for conn in proc.net_connections(kind="inet"):
                    if conn.status == psutil.CONN_LISTEN:
                        # TODO: Are we considering IPv6 or assuming that it will always be available over IPv4?
                        laddr = (
                            f"{conn.laddr.ip}:{conn.laddr.port}"
                            if conn.laddr.ip != "::"
                            else f"127.0.0.1:{conn.laddr.port}"
                        )
                        self.listening_on = f"http://{laddr}"
                        # Just take the first listening address
                        break
            except psutil.NoSuchProcess:
                ...
            except psutil.AccessDenied:
                logger.warning(
                    f"Parent process ID of process {proc.name()} ({self.process_id}, {proc.status()}) cannot be retrieved. Run auto-config as super-user."
                )
        return self.listening_on

    def infer_models_dir_path(self) -> str:
        """Get the path to the models directory used by Ollama."""
        self.models_dir_path = self.process_env_vars.get("OLLAMA_MODELS", "")
        if self.models_dir_path == "":
            # raise NotImplementedError("This method has been partly implemented only.")
            if self.infer_listening_on() != "":
                ollama_client = OllamaClient(host=self.listening_on)
                models_list = ollama_client.list().models
                if len(models_list) > 0:
                    modelfile_text = ollama_client.show(models_list[0].model).modelfile
                    pattern = re.compile(
                        r"^\s*FROM\s+(.+?)(?:\s*#.*)?$", re.IGNORECASE | re.MULTILINE
                    )
                    match = pattern.search(modelfile_text)
                    if match:
                        model_blob_path = match.group(1).strip()
                        likely_models_dir = str(
                            os.path.dirname(os.path.dirname(model_blob_path))
                        )
                        self.models_dir_path = likely_models_dir.replace(
                            os.path.expanduser("~"), "~"
                        )
                else:
                    logger.warning(
                        "No models are currently installed in Ollama. Cannot infer the models directory path."
                    )
        return self.models_dir_path

    def is_likely_daemon(self) -> bool:
        """Infer if the Ollama process is likely running as a daemon/service."""
        self.get_parent_process_id()
        if self.parent_process_id not in [-1, 1]:
            self.likely_daemon = False
        else:
            proc = psutil.Process(self.process_id)
            if (
                proc.username().lower() in ["ollama", "root"]
                and proc.terminal() is None
            ):
                self.likely_daemon = True
            else:
                self.likely_daemon = False
        return self.likely_daemon



from ollama_downloader.data.data_models import AppSettings, OllamaLibrary, OllamaServer


class TestDataModels:
    """
    Class to group tests related to data models.
    """

    def test_app_settings(self):
        """
        Test the AppSettings data model.
        """
        settings = AppSettings.load_or_create_default()
        assert settings is not None
        assert isinstance(settings.ollama_server, OllamaServer)
        assert isinstance(settings.ollama_library, OllamaLibrary)

        another_settings = AppSettings()
        yet_another_settings = AppSettings()
        # Singleton behavior
        assert id(settings) == id(another_settings) == id(yet_another_settings)

        another_settings.ollama_library.verify_ssl = False
        # Singleton behavior check: changing one instance changes all
        assert settings.ollama_library.verify_ssl is False

    def test_app_settings_invalid_urls(self):
        """
        Test the AppSettings data model parts with invalid URLs.
        """
        invalid_urls = [
            "htp://invalid-url",  # Invalid scheme
            "http://",  # Missing host
            "http://invalid-url:port",  # Invalid port
            "http://invalid-url/path with spaces",  # Spaces in URL
            "http://inva|id-url",  # Invalid character
        ]

        for url in invalid_urls:
            try:
                OllamaServer(url=url)
                assert False, f"Expected validation error for URL: {url}"
            except Exception as e:
                assert "ValidationError" in str(type(e)) or "AssertionError" in str(
                    type(e)
                )

            try:
                OllamaLibrary(library_base_url=url, registry_base_url=url)
                assert False, f"Expected validation error for URL: {url}"
            except Exception as e:
                assert "ValidationError" in str(type(e)) or "AssertionError" in str(
                    type(e)
                )

    def test_app_settings_invalid_path(self):
        """
        Test the OllamaLibrary data model with an invalid models_path.
        """
        invalid_paths = [
            "/path/that/does/not/exist",
            "/etc/passwd",  # Not a directory
            "https://not.a.path",  # Not a directory path
        ]

        for path in invalid_paths:
            try:
                OllamaLibrary(models_path=path)
                assert False, f"Expected validation error for path: {path}"
            except Exception as e:
                assert "ValidationError" in str(type(e)) or "AssertionError" in str(
                    type(e)
                )



from typer.testing import CliRunner

import pytest

from ollama_downloader.cli import app
from ollama_downloader.data.data_models import AppSettings

from importlib.metadata import version


class TestTyperCalls:
    """
    Class to group tests related to Typer CLI commands.
    """

    PACKAGE_NAME = "ollama-downloader"

    @pytest.fixture(autouse=True)
    def runner(self):
        """
        Fixture to provide a Typer CLI runner for testing.
        """
        runner = CliRunner()
        return runner

    def test_version(self, runner):
        """
        Test the 'version' command of the CLI.
        """
        result = runner.invoke(app=app, args=["version"])
        expected_version = version(self.PACKAGE_NAME)
        assert expected_version in result.output
        assert result.exit_code == 0

    def test_show_config(self, runner):
        """
        Test the 'show-config' command of the CLI.
        """
        result = runner.invoke(app=app, args=["show-config"])
        assert result.exit_code == 0
        settings = AppSettings.load_settings()
        # Assert that we can read the local settings -- after all, this is what the show-config command does
        assert settings is not None
        # This is a bit fragile, as the indentation matching depends on the print_json implementation of Rich, which defaults to 2.
        assert settings.model_dump_json(indent=2) == result.output.strip()

    def test_auto_config(self, runner):
        """
        Test the 'auto-config' command of the CLI.
        """
        result = runner.invoke(app=app, args=["auto-config"])
        assert result.exit_code == 0
        if result.output.strip() != "":
            assert AppSettings.model_validate_json(result.output.strip()) is not None

    def test_list_models(self, runner):
        """
        Test the 'list-models' command of the CLI.
        """
        result = runner.invoke(app=app, args=["list-models"])
        assert result.exit_code == 0
        # Expect at least few known models to be listed
        assert "gpt-oss" in result.output.lower()
        assert "llama" in result.output.lower()
        assert "granite" in result.output.lower()
        assert "gemma" in result.output.lower()
        assert "deepseek" in result.output.lower()
        assert "made-up-model-that-should-not-exist" not in result.output.lower()

    def test_list_tags(self, runner):
        """
        Test the 'list-tags' command of the CLI.
        """
        result = runner.invoke(app, ["list-tags", model_identifier := "gpt-oss"])
        assert result.exit_code == 0
        # Expect at least two known tags and a cloud tag to be listed for the gpt-oss model
        assert f"{model_identifier}:latest" in result.output
        assert f"{model_identifier}:20b" in result.output
        assert f"{model_identifier}:20b-cloud" in result.output
        result = runner.invoke(
            app=app,
            args=["list-tags", "made-up-model-that-should-not-exist"],
        )
        # Should be an empty output while the error will be logged but exit code will still be 0
        assert result.output == ""

    def test_model_download(self, runner):
        """
        Test the 'model-download' command of the CLI.
        """
        # Let's try downloading the smallest possible model to stop the test from taking too long
        model_tag = "all-minilm:22m"
        result = runner.invoke(app=app, args=["model-download", model_tag])
        assert result.exit_code == 0
        assert f"{model_tag} successfully downloaded and saved" in result.output

        # Let's try downloading a cloud model
        model_tag = "gpt-oss:20b-cloud"
        result = runner.invoke(app=app, args=["model-download", model_tag])
        assert result.exit_code == 0
        assert f"{model_tag} successfully downloaded and saved" in result.output

        model_tag = "made-up:should-fail"
        result = runner.invoke(app=app, args=["model-download", model_tag])
        assert result.exit_code == 0
        assert f"{model_tag} successfully downloaded and saved" not in result.output

    def test_hf_list_models(self, runner):
        """
        Test the hf-list-models' command of the CLI.
        """
        result = runner.invoke(app=app, args=["hf-list-models", "--page", "4"])
        assert result.exit_code == 0
        # Expect the output to contain at least 25 models on page 4
        # Models change often on Hugging Face, so we cannot check for specific models
        assert "25, page 4" in result.output.lower()
        assert "made-up-model-that-should-not-exist" not in result.output.lower()

    def test_hf_list_tags(self, runner):
        """
        Test the 'list-tags' command of the CLI.
        """
        result = runner.invoke(
            app,
            ["hf-list-tags", model_identifier := "unsloth/SmolLM2-135M-Instruct-GGUF"],
        )
        assert result.exit_code == 0
        # Expect at least two known tags to be listed for the gpt-oss model
        assert f"{model_identifier}:F16" in result.output
        assert f"{model_identifier}:Q4_K_M" in result.output
        result = runner.invoke(
            app=app,
            args=["hf-list-tags", "made-up-model-that-should-not-exist"],
        )
        # Should be an empty output while the error will be logged but exit code will still be 0
        assert result.output == ""

    def test_hf_model_download(self, runner):
        """
        Test the 'hf-model-download' command of the CLI.
        """
        # Let's try downloading the smallest possible model to stop the test from taking too long
        user_repo_quant = "unsloth/SmolLM2-135M-Instruct-GGUF:Q4_K_M"
        result = runner.invoke(
            app=app,
            args=["hf-model-download", user_repo_quant],
        )
        assert result.exit_code == 0
        assert f"{user_repo_quant} successfully downloaded and saved" in result.output

        user_repo_quant = "made-up/should-fail:Q0_0_X"
        result = runner.invoke(
            app=app,
            args=["hf-model-download", user_repo_quant],
        )
        assert result.exit_code == 0
        assert (
            f"{user_repo_quant} successfully downloaded and saved" not in result.output
        )



from abc import ABC, abstractmethod
from enum import Enum
import hashlib
import logging
import os
import platform
import shutil
import ssl
import tempfile
from typing import List, Set, Tuple
from urllib.parse import urlparse

from ollama import Client as OllamaClient

import certifi
import httpx

from environs import env

from ollama_downloader import EnvVar

from rich.progress import (
    Progress,
    BarColumn,
    TextColumn,
    DownloadColumn,
    TransferSpeedColumn,
)

from ollama_downloader.data.data_models import AppSettings

logger = logging.getLogger(__name__)


class ModelSource(Enum):
    OLLAMA = 1
    HUGGINGFACE = 2


class ModelDownloader(ABC):
    HF_BASE_URL = "https://hf.co/v2/"

    def __init__(self):
        self.settings: AppSettings = AppSettings.load_or_create_default()
        if not self.settings:
            raise RuntimeError("Failed to load or create and save default settings.")
        self._user_agent: str = f"{EnvVar.OD_UA_NAME_VER} ({platform.platform()} {platform.system()}-{platform.release()} Python-{platform.python_version()})"
        self._unnecessary_files: Set[str] = set()
        self._cleanup_running: bool = False

    @abstractmethod
    def download_model(self, model_identifier: str) -> bool:
        """
        Download a supported model into an available Ollama server.

        Args:
            model_identifier (str): The model tag to download, e.g., "gpt-oss:latest" for library models.
            If the tag is omitted, "latest" is assumed. For Hugging Face models, the model identifier is
            of the format <user>/<repository>:<quantisation>, e.g., unsloth/gemma-3-270m-it-GGUF:Q4_K_M.

        Returns:
            bool: True if the model was successfully downloaded and verified, False otherwise.
        """
        pass

    @abstractmethod
    def list_available_models(
        self, page: int | None = None, page_size: int | None = None
    ) -> List[str]:
        """
        List available models. If pagination is supported by the source, page and page_size can be used to control the results.
        If pagination is not supported or the page and page_size are None, the number of models will be returned will depend on
        the implementing sub-class.

        Args:
            page (int | None): The page number to retrieve, if pagination is supported.
            page_size (int | None): The number of items per page, if pagination is supported

        Returns:
            list[str]: A list of available model identifiers, excluding any of their tags.
        """
        pass

    @abstractmethod
    def list_model_tags(self, model_identifier: str) -> List[str]:
        """
        List available tags for a specific model.

        Args:
            model_identifier (str): The name of the model to list tags for, e.g., "gpt-oss" for an Ollama library model
            or "unsloth/gemma-3-270m-it-GGUF" for a Hugging Face model.

        Returns:
            list[str]: A list of available tags for the specified model, e.g., ["latest", "20b"] or
            the Hugging Face quantisations such as ["Q4_K_M", "Q4_K_S"].
        """
        pass

    def get_httpx_client(self, verify: bool, timeout: float) -> httpx.Client:
        """
        Obtain an HTTPX client for making requests.

        Args:
            verify (bool): Whether to verify SSL certificates.
            timeout (float): The timeout for requests in seconds.

        Returns:
            httpx.Client: An HTTPX client configured with the specified settings.
        """
        if verify is False:
            logger.warning(
                "SSL verification is disabled. This is not recommended for production use."
            )
        ctx = ssl.create_default_context(
            cafile=env.str("SSL_CERT_FILE", default=certifi.where()),
            capath=env.str("SSL_CERT_DIR", default=None),
        )
        client = httpx.Client(
            verify=verify if (verify is not None and verify is False) else ctx,
            follow_redirects=True,
            trust_env=True,
            http2=True,
            timeout=timeout,
            headers={"User-Agent": self._user_agent},
        )
        return client

    def _make_manifest_url(
        self, model_identifier: str, model_source: ModelSource
    ) -> httpx.URL:
        """
        Constructs the manifest URL for a given model identifier.

        Args:
            model_identifier (str): The model identifier, e.g., "gpt-oss:latest" for an Ollama library model
            or "unsloth/gemma-3-270m-it-GGUF:Q4_K_M" for a Hugging Face model.
            model_source (ModelSource): The source of the model (e.g., OLLAMA or HUGGINGFACE).

        Returns:
            httpx.URL: The constructed manifest URL.
        """
        model, tag = (
            model_identifier.split(":")
            if ":" in model_identifier
            else (model_identifier, "latest")
        )
        match model_source:
            case ModelSource.OLLAMA:
                logger.debug(f"Constructing manifest URL for {model}:{tag}")
                return httpx.URL(self.settings.ollama_library.registry_base_url).join(
                    f"{model}/manifests/{tag}"
                )
            case ModelSource.HUGGINGFACE:
                logger.debug(f"Constructing manifest URL for {model_identifier}")
                hf_model_identifier = (
                    model_identifier.replace(":", "/manifests/")
                    if ":" in model_identifier
                    else f"{model_identifier}/manifests/{tag}"
                )
                return httpx.URL(f"{ModelDownloader.HF_BASE_URL}{hf_model_identifier}")
            case _:
                raise ValueError(f"Unsupported model source: {model_source}")

    def _fetch_manifest(self, model_identifier: str, model_source: ModelSource) -> str:
        """
        Fetches the manifest JSON content for a given model identifier.

        Args:
            model_identifier (str): The model identifier, e.g., "gpt-oss:latest" for an Ollama library model
            or "unsloth/gemma-3-270m-it-GGUF:Q4_K_M" for a Hugging Face model.
            model_source (ModelSource): The source of the model (e.g., OLLAMA or HUGGINGFACE).

        Returns:
            str: The fetched manifest JSON content as a string.
        """
        url = self._make_manifest_url(model_identifier, model_source)
        logger.info(f"Downloading manifest from {url}")
        with self.get_httpx_client(
            self.settings.ollama_library.verify_ssl,
            self.settings.ollama_library.timeout,
        ) as http_client:
            response = http_client.get(url)
            response.raise_for_status()
            return response.text

    def _make_blob_url(
        self, model_identifier: str, digest: str, model_source: ModelSource
    ) -> httpx.URL:
        """
        Constructs the blob URL for a given model and digest.

        Args:
            model_identifier (str): The model name, e.g., "gpt-oss:latest".
            digest (str): The digest of the blob, e.g., "sha256:abcdef...".
            model_source (ModelSource): The source of the model (e.g., OLLAMA or HUGGINGFACE).

        Returns:
            httpx.URL: The constructed blob URL.
        """
        logger.debug(
            f"Constructing blob URL for {model_identifier} with digest {digest}"
        )
        model_name = model_identifier.split(":")[0]
        match model_source:
            case ModelSource.OLLAMA:
                return httpx.URL(self.settings.ollama_library.registry_base_url).join(
                    f"{model_name}/blobs/{digest.replace(':', '-')}"
                )
            case ModelSource.HUGGINGFACE:
                return httpx.URL(
                    f"{ModelDownloader.HF_BASE_URL}{model_name}/blobs/{digest}"
                )
            case _:
                raise ValueError(f"Unsupported model source: {model_source}")

    def _download_model_blob(
        self, model_identifier: str, named_digest: str, model_source: ModelSource
    ) -> tuple:
        """
        Downloads a model blob given its named digest.

        Args:
            model_identifier (str): The model name, e.g., "gpt-oss:latest".
            named_digest (str): The named digest of the blob, e.g., "sha256:abcdef...".
            model_source (ModelSource): The source of the model (e.g., OLLAMA or HUGGINGFACE).

        Returns:
            tuple: A tuple containing the path to the downloaded file as string and its digest.
        """
        url = self._make_blob_url(
            model_identifier=model_identifier,
            digest=named_digest,
            model_source=model_source,
        )
        # try:
        sha256_hash = hashlib.new("sha256")
        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            self._unnecessary_files.add(temp_file.name)
            with self.get_httpx_client(
                self.settings.ollama_library.verify_ssl,
                self.settings.ollama_library.timeout,
            ).stream("GET", url) as response:
                response.raise_for_status()
                total = int(response.headers["Content-Length"])

                with Progress(
                    TextColumn(text_format="{task.description}"),
                    "[progress.percentage]{task.percentage:>3.0f}%",
                    BarColumn(bar_width=None),
                    DownloadColumn(),
                    TransferSpeedColumn(),
                ) as progress:
                    download_task = progress.add_task(
                        f"Downloading BLOB {named_digest[:11]}...{named_digest[-4:]}",
                        total=total,
                    )
                    for chunk in response.iter_bytes():
                        sha256_hash.update(chunk)
                        temp_file.write(chunk)
                        progress.update(
                            download_task, completed=response.num_bytes_downloaded
                        )
        logger.debug(f"Downloaded {url} to {temp_file.name}")
        content_digest = sha256_hash.hexdigest()
        logger.debug(f"Computed SHA256 digest of {temp_file.name}: {content_digest}")
        return temp_file.name, content_digest

    def _save_manifest(
        self, data: str, model_identifier: str, model_source: ModelSource
    ):
        """
        Saves the manifest data to the appropriate location based on the model source.

        Args:
            data (str): The manifest JSON content as a string.
            model_identifier (str): The model identifier, e.g., "gpt-oss:latest" for an Ollama library model
            or "unsloth/gemma-3-270m-it-GGUF:Q4_K_M" for a Hugging Face model.
            model_source (ModelSource): The source of the model (e.g., OLLAMA or HUGGINGFACE).

        Returns:
            str: The path to the saved manifest file.
        """
        manifests_toplevel_dir = os.path.join(
            (
                os.path.expanduser(self.settings.ollama_library.models_path)
                if self.settings.ollama_library.models_path.startswith("~")
                else self.settings.ollama_library.models_path
            ),
            "manifests",
        )
        model_identifier_splits = model_identifier.split(":")
        manifests_dir = None
        model_source_host = None
        match model_source:
            case ModelSource.OLLAMA:
                model_source_host = urlparse(
                    self.settings.ollama_library.registry_base_url
                ).hostname
                manifests_dir = os.path.join(
                    manifests_toplevel_dir,
                    model_source_host,
                    "library",
                    model_identifier_splits[0],
                )
            case ModelSource.HUGGINGFACE:
                model_source_host = urlparse(ModelDownloader.HF_BASE_URL).hostname
                manifests_dir = os.path.join(
                    manifests_toplevel_dir,
                    # FIXME: This is a hack for mypy -- the model_source_host should not be None
                    model_source_host or "",
                    model_identifier_splits[0],
                )
            case _:
                raise ValueError(f"Unsupported model source: {model_source}")
        if not os.path.exists(manifests_dir):
            logger.warning(
                f"Manifests path {manifests_dir} does not exist. Will attempt to create it."
            )
            os.makedirs(manifests_dir)
            self._unnecessary_files.add(manifests_dir)
        target_file = os.path.join(manifests_dir, model_identifier_splits[1])
        with open(target_file, "w") as f:
            f.write(data)
            logger.info(f"Saved manifest to {target_file}")
        if self.settings.ollama_library.user_group:
            user, group = self.settings.ollama_library.user_group
            shutil.chown(target_file, user, group)
            # The directory ownership must also be changed because it may have been created by a different user, most likely a sudoer
            # TODO: Is this necessary or can the ownership change to the top-level directory cascade down?
            shutil.chown(manifests_dir, user, group)
            shutil.chown(manifests_toplevel_dir, user, group)
            logger.info(
                f"Changed ownership of {target_file} to user: {user}, group: {group}"
            )
        self._unnecessary_files.add(target_file)
        return target_file

    def _save_blob(
        self,
        source: str,
        named_digest: str,
        computed_digest: str,
    ) -> Tuple[bool, str | None]:
        """
        Saves the downloaded blob to the appropriate location based on the model source.

        Args:
            source (str): The path to the downloaded BLOB.
            named_digest (str): The expected digest of the BLOB prefixed with the digest algorithm followed by the colon character.
            computed_digest (str): The computed digest of the BLOB.

        Returns:
            Tuple[bool, str | None]: A tuple containing a boolean indicating success or failure,
            and the path to the saved BLOB if successful, None otherwise.
        """
        if computed_digest != named_digest[7:]:
            logger.error(
                f"Digest mismatch: expected {named_digest[7:]}, got {computed_digest}"
            )
            return False, None

        blobs_dir = os.path.join(
            (
                os.path.expanduser(self.settings.ollama_library.models_path)
                if self.settings.ollama_library.models_path.startswith("~")
                else self.settings.ollama_library.models_path
            ),
            "blobs",
        )

        logger.info(f"BLOB {named_digest} digest verified successfully.")
        if not os.path.isdir(blobs_dir):
            logger.error(f"BLOBS path {blobs_dir} must be a directory.")
            return False, None
        if not os.path.exists(blobs_dir):
            logger.error(f"BLOBS path {blobs_dir} must exist.")
            return False, None
        target_file = os.path.join(blobs_dir, named_digest.replace(":", "-"))
        shutil.move(source, target_file)
        self._unnecessary_files.remove(source)
        self._unnecessary_files.add(target_file)
        logger.info(f"Moved {source} to {target_file}")
        if self.settings.ollama_library.user_group:
            user, group = self.settings.ollama_library.user_group
            shutil.chown(target_file, user, group)
            shutil.chown(blobs_dir, user, group)
            # Set permissions to rw-r-----
            os.chmod(target_file, 0o640)
            logger.info(
                f"Changed ownership of {target_file} to user: {user}, group: {group}"
            )
        return True, target_file

    def cleanup_unnecessary_files(self):
        """
        Cleans up unnecessary files and directories created during downloading models.
        """
        # TODO: Is this thread-safe? Should we use a lock?
        if not self._cleanup_running:
            self._cleanup_running = True
            list_of_unnecessary_files = list(self._unnecessary_files)
            unnecessary_directories = set()
            for file_object in list_of_unnecessary_files:
                try:
                    if not os.path.isdir(file_object):
                        os.remove(file_object)
                        logger.info(f"Removed unnecessary file: {file_object}")
                    else:
                        # If it's a directory, we don't remove it yet because it may not be empty.
                        unnecessary_directories.add(file_object)
                    self._unnecessary_files.remove(file_object)
                except Exception as e:
                    logger.error(
                        f"Failed to remove unnecessary file {file_object}: {e}"
                    )

            # Now remove unnecessary directories if they are empty
            for directory in unnecessary_directories:
                try:
                    os.rmdir(directory)
                    logger.info(f"Removed unnecessary directory: {directory}")
                except OSError as e:
                    logger.error(
                        f"Failed to remove unnecessary directory {directory}: {e}"
                    )
            self._cleanup_running = False

    def remove_model(self, model_identifier: str) -> bool:  # pragma: no cover
        """
        Removes a model from the Ollama server.
        """
        ollama_client = OllamaClient(
            host=self.settings.ollama_server.url,
            # timeout=self.settings.ollama_server.timeout,
            # TODO: Add API key authentication logic
        )
        search_model = (
            f"{urlparse(ModelDownloader.HF_BASE_URL).hostname}/{model_identifier}"
        )
        response = ollama_client.delete(search_model)
        if hasattr(response, "status") and response.status == "success":
            logger.info(
                f"Successfully removed model {model_identifier} from Ollama server at {self.settings.ollama_server.url}."
            )
            return True
        else:
            logger.error(
                f"Failed to remove model {model_identifier} from Ollama server at {self.settings.ollama_server.url}."
            )
            return False



import os
import httpx
from ollama_downloader.sysinfo import OllamaSystemInfo


class TestOllamaSystemInfo:
    """
    Class to group tests related to the OllamaSystemInfo class.
    Assume that Ollama is NOT running as a service/daemon for these tests.
    """

    def test_is_running(self):
        system_info = OllamaSystemInfo()
        assert system_info.is_running() is True

    def test_is_not_windows(self):
        system_info = OllamaSystemInfo()
        assert system_info.is_windows() is False

    def test_access_env_vars(self):
        system_info = OllamaSystemInfo()
        assert isinstance(system_info.process_env_vars, dict)
        # Assuming that the environment variable "PATH" is set in the Ollama process
        assert "PATH" in system_info.process_env_vars
        # assert "OLLAMA_MODELS" in system_info.process_env_vars

    def test_process_owner(self):
        system_info = OllamaSystemInfo()
        owner_info = system_info.get_process_owner()
        assert isinstance(owner_info, tuple)
        assert len(owner_info) == 4  # (username, uid, groupname, gid)
        assert isinstance(owner_info[0], str)  # username
        assert owner_info[0] != ""  # username should not be empty
        assert isinstance(owner_info[1], int)  # uid
        assert owner_info[1] > 0  # uid should be a positive integer
        assert isinstance(owner_info[2], str)  # groupname
        assert owner_info[2] != ""  # groupname should not be empty
        assert isinstance(owner_info[3], int)  # gid
        assert owner_info[3] > 0  # gid should be a positive integer
        assert (
            owner_info[0] != "ollama" and owner_info[2] != "ollama"
        )  # Assuming Ollama is not running as user/group "ollama"

    def test_get_parent_process_id(self):
        system_info = OllamaSystemInfo()
        parent_pid = system_info.get_parent_process_id()
        assert isinstance(parent_pid, int)
        assert parent_pid > 0  # Parent PID should be a positive integer

    def test_infer_listening_on(self):
        system_info = OllamaSystemInfo()
        listening_on = system_info.infer_listening_on()
        assert listening_on != ""
        # Try connecting and check the response
        with httpx.Client(timeout=30.0) as client:
            response = client.get(listening_on)
            assert response.status_code == 200
            assert response.text == "Ollama is running"

    # @pytest.mark.skip(reason="Disabled temporarily.")
    def test_is_model_dir_env_var_set(self):
        system_info = OllamaSystemInfo()
        # Assuming that the environment variable "OLLAMA_MODELS" has not been passed to the Ollama process
        assert system_info.is_model_dir_env_var_set() is False

    # @pytest.mark.skip(reason="This feature is currently under development.")
    def test_is_likely_daemon(self):
        system_info = OllamaSystemInfo()
        assert system_info.is_likely_daemon() is False

    # @pytest.mark.skip(reason="This feature is currently under development.")
    def test_infer_models_dir_path(self):
        system_info = OllamaSystemInfo()
        models_path = os.path.expanduser(system_info.infer_models_dir_path())
        assert models_path is not None
        if models_path != "":
            # Check that the inferred models path exists and is a directory
            assert os.path.exists(models_path)
            assert os.path.isdir(models_path)



import asyncio
import logging
import signal
import sys
from types import FrameType
from typing import Optional
from httpx import HTTPStatusError
from typing_extensions import Annotated
import typer
from rich import print as print
from rich import print_json
import psutil
from importlib.metadata import version as metadata_version

from ollama_downloader.sysinfo import OllamaSystemInfo
from ollama_downloader.data.data_models import AppSettings
from ollama_downloader.downloader.ollama_model_downloader import OllamaModelDownloader
from ollama_downloader.downloader.hf_model_downloader import HuggingFaceModelDownloader


# Initialize the logger
logger = logging.getLogger(__name__)

# Initialize the Typer application
app = typer.Typer(
    no_args_is_help=True,
    add_completion=False,
    help="A command-line interface for the Ollama downloader.",
)


class OllamaDownloaderCLIApp:
    def __init__(self):
        # Set up signal handlers for graceful shutdown
        for sig in [signal.SIGINT, signal.SIGTERM]:
            signal.signal(sig, self._interrupt_handler)
        self._model_downloader: OllamaModelDownloader = None
        self._hf_model_downloader: HuggingFaceModelDownloader = None

    def _interrupt_handler(
        self, signum: int, frame: FrameType | None
    ):  # pragma: no cover
        logger.warning("Interrupt signal received, performing clean shutdown")
        logger.debug(f"Interrupt signal number: {signum}. Frame: {frame}")
        # Cleanup will be performed due to the finally block in each command
        sys.exit(0)

    def _initialize(self):
        logger.debug("Initializing downloaders...")
        if not self._model_downloader:
            self._model_downloader = OllamaModelDownloader()
        if not self._hf_model_downloader:
            self._hf_model_downloader = HuggingFaceModelDownloader()

    def _cleanup(self):
        logger.debug("Running cleanup...")

        if self._model_downloader:
            self._model_downloader.cleanup_unnecessary_files()
        if self._hf_model_downloader:
            self._hf_model_downloader.cleanup_unnecessary_files()

        logger.debug("Cleanup completed.")

    async def _version(self):
        package_name = "ollama-downloader"
        name_splits = package_name.split("-")
        if len(name_splits) != 2:
            abbreviation = package_name
        else:
            abbreviation = f"{name_splits[0][0]}{name_splits[1][0]}"
        return (
            f"{package_name} ({abbreviation}) version {metadata_version(package_name)}"
        )

    async def run_version(self):
        try:
            result = await self._version()
            print(result)
        except Exception as e:
            logger.error(f"Error in getting version. {e}")

    async def _show_config(self):
        return self._model_downloader.settings.model_dump_json()

    async def run_show_config(self):
        try:
            self._initialize()
            result = await self._show_config()
            print_json(json=result)
        except Exception as e:
            logger.error(f"Error in showing config. {e}")
        finally:
            self._cleanup()

    async def _auto_config(self):
        logger.warning(
            "Automatic configuration is an experimental feature. Its output maybe incorrect!"
        )
        system_info = OllamaSystemInfo()
        if system_info.is_windows():
            raise NotImplementedError(
                "Automatic configuration is not supported on Windows yet."
            )
        super_user_needed = False
        super_user_needed = super_user_needed or system_info.infer_listening_on() in [
            None,
            "",
        ]
        super_user_needed = (
            super_user_needed or system_info.infer_models_dir_path() in [None, ""]
        )
        if super_user_needed:
            return {}
        inferred_settings = AppSettings()
        inferred_settings.ollama_server.url = system_info.listening_on
        inferred_settings.ollama_library.models_path = system_info.models_dir_path
        if system_info.is_likely_daemon():
            if system_info.is_macos():
                logger.warning(
                    "Automatic configuration on macOS maybe flawed if Ollama is configured to run as a background service."
                )
            inferred_settings.ollama_library.user_group = (
                system_info.process_owner[0],
                system_info.process_owner[2],
            )
        return inferred_settings.model_dump_json()

    # async def old_auto_config(self):
    #     logger.warning(
    #         "Automatic configuration is a future experimental idea, which has not been implemented yet. Stay tuned!"
    #     )
    #     logger.warning(
    #         "Some relevant information about Ollama will be displayed but these are based on assumptions that may not always be true."
    #     )
    #     # TODO: The following stuff must go in its own module/class with its own tests. This is just a placeholder.
    #     # The idea is to gather relevant information about the Ollama installation and use it to infer a configuration.
    #     # 1. Check if Ollama is installed and running.
    #     # 2. Output the IP/interface and port the Ollama server is listening on.
    #     # 3. Output the Ollama executable path.
    #     # 4. Output user/group Ollama is running as.
    #     # 5. Attempt to connect to the running Ollama server and try to infer the model storage path from the first modelfile.

    #     if platform.system() == "Windows":
    #         # Don't bother going there until we have a real Windows machine to test on.
    #         raise NotImplementedError(
    #             "Automatic configuration is not supported on Windows yet. Probably never will be!"
    #         )
    #     relevant_info: Dict[str, Any] = {}

    #     pid: int = -1
    #     for proc in psutil.process_iter(["pid", "name"]):
    #         if proc.info["name"] == "ollama":
    #             pid = proc.info["pid"]
    #     if pid < 0:
    #         raise RuntimeError(
    #             "Ollama process not found. Is Ollama installed and running?"
    #         )
    #     process = psutil.Process(pid)
    #     relevant_info["process"] = {"name": process.name(), "pid": pid}
    #     parent_id = process.ppid()
    #     relevant_info["parent"] = {
    #         "name": psutil.Process(parent_id).name(),
    #         "pid": parent_id,
    #     }
    #     # User and Group Information
    #     username = process.username()

    #     owner = {}
    #     effective_uid = process.uids().effective
    #     effective_gid = process.gids().effective
    #     owner["user"] = {"name": username, "uid": effective_uid}
    #     owner["group"] = {
    #         "name": grp.getgrgid(effective_gid).gr_name,
    #         "gid": effective_gid,
    #     }
    #     relevant_info["owner"] = owner

    #     relevant_info["sudo_needed_to_download_models"] = os.getuid() != effective_uid

    #     relevant_info["executable"] = process.exe()
    #     relevant_info["cmdline"] = process.cmdline()

    #     explicit_models_dir = process.environ().get("OLLAMA_MODELS", None)
    #     relevant_info["model_dir_explicitly_specified"] = (
    #         explicit_models_dir is not None
    #     )

    #     relevant_info["http_proxy_specified"] = (
    #         process.environ().get("HTTP_PROXY", None) is not None
    #         or process.environ().get("http_proxy", None) is not None
    #         or process.environ().get("HTTPS_PROXY", None) is not None
    #         or process.environ().get("https_proxy", None) is not None
    #     )

    #     listening_on = []
    #     ollama_url = ""
    #     for conn in process.net_connections(kind="inet"):
    #         if conn.status == psutil.CONN_LISTEN:
    #             laddr = (
    #                 f"{conn.laddr.ip}:{conn.laddr.port}"
    #                 if conn.laddr.ip != "::"
    #                 else f"127.0.0.1:{conn.laddr.port}"
    #             )
    #             listening_on.append(laddr)
    #             ollama_url = f"http://{laddr}"
    #             # Just take the first listening address
    #             break
    #     relevant_info["listening_on"] = listening_on

    #     # FIXME: This is not a good idea, since we are including a specific check
    #     # but most systems use this init system so it maybe fine.
    #     systemd_service = "/etc/systemd/system/ollama.service"
    #     systemd_daemon = os.path.exists(systemd_service)

    #     ollama_client = OllamaClient(host=ollama_url)
    #     models_list = ollama_client.list().models
    #     if explicit_models_dir is not None:
    #         relevant_info["likely_models_path"] = explicit_models_dir
    #     elif systemd_daemon:
    #         # https://ollama.com/install.sh :: function configure_systemd()
    #         # if systemd is the init system we use the /usr/share/ollama directory for all installs
    #         relevant_info["likely_models_path"] = "/usr/share/ollama/.ollama/models"
    #     elif len(models_list) > 0:
    #         modelfile_text = ollama_client.show(models_list[0].model).modelfile
    #         pattern = re.compile(
    #             r"^\s*FROM\s+(.+?)(?:\s*#.*)?$", re.IGNORECASE | re.MULTILINE
    #         )
    #         match = pattern.search(modelfile_text)
    #         if match:
    #             model_blob_path = match.group(1).strip()
    #             likely_models_dir = str(
    #                 os.path.dirname(os.path.dirname(model_blob_path))
    #             )
    #             relevant_info["likely_models_path"] = likely_models_dir.replace(
    #                 os.path.expanduser("~"), "~"
    #             )
    #             real_models_path = os.path.realpath(likely_models_dir)
    #             is_symlink_in_path = real_models_path != likely_models_dir
    #             if is_symlink_in_path:
    #                 relevant_info["symlink_in_models_path"] = is_symlink_in_path
    #                 relevant_info["likely_real_models_path"] = real_models_path
    #                 stat_info = os.stat(real_models_path)
    #                 relevant_info["likely_real_models_path_owner"] = {
    #                     "user": {
    #                         "name": pwd.getpwuid(stat_info.st_uid).pw_name,
    #                         "uid": stat_info.st_uid,
    #                     },
    #                     "group": {
    #                         "name": grp.getgrgid(stat_info.st_gid).gr_name,
    #                         "gid": stat_info.st_gid,
    #                     },
    #                 }
    #     else:
    #         # We could select ~/.ollama/models as a default directory
    #         relevant_info["likely_models_path"] = "~/.ollama/models"
    #         logger.warning(
    #             "No models found in the running Ollama instance. Models path cannot be computed. Using defaults."
    #         )

    #     open_files = []
    #     for file in process.open_files():
    #         open_files.append(file.path)
    #     relevant_info["ollama_open_files"] = open_files

    #     # auto-config does not need to fail if we can't detect if ollama is a daemon or not
    #     try:
    #         relevant_info["ollama_is_likely_daemon"] = systemd_daemon or (
    #             process.terminal() is None
    #             and process.status()
    #             not in [psutil.STATUS_RUNNING, psutil.STATUS_SLEEPING]
    #             and process.ppid() != 1
    #         )
    #     except Exception as e:
    #         if isinstance(e, psutil.AccessDenied):
    #             logger.info(
    #                 "Seems like you need to run this command with super-user permissions. Try `sudo`!"
    #             )

    #     current_settings = self._model_downloader.settings

    #     new_settings = current_settings.model_dump()
    #     new_settings["ollama_library"]["user_group"] = (
    #         owner["user"]["name"],
    #         owner["group"]["name"],
    #     )
    #     new_settings["ollama_library"]["models_path"] = relevant_info[
    #         "likely_models_path"
    #     ]

    #     for entry in new_settings["ollama_library"]:
    #         value = new_settings["ollama_library"][entry]
    #         current_value = current_settings.ollama_library.__dict__[entry]
    #         if current_value != value:
    #             logger.info(
    #                 f"Change value of {entry} from {current_value} to {value}. Updating the `conf/settings.json` may be nessesary."
    #             )

    #     # we do not update anything without the user consent
    #     return current_settings.model_dump_json()

    async def run_auto_config(self):
        try:
            # self._initialize()
            result = await self._auto_config()
            if result != {}:
                print_json(json=result)
        except Exception as e:
            logger.error(f"Error in generating automatic config. {e}")
            if isinstance(e, psutil.AccessDenied):
                logger.info(
                    "Seems like you need to run this command with super-user permissions. Try `sudo`!"
                )
        finally:
            self._cleanup()

    async def _list_models(self, page: int | None = None, page_size: int | None = None):
        return self._model_downloader.list_available_models(
            page=page, page_size=page_size
        )

    async def run_list_models(
        self, page: int | None = None, page_size: int | None = None
    ):
        try:
            self._initialize()
            result = await self._list_models(page=page, page_size=page_size)
            if page and page_size and page_size >= len(result):
                print(f"Model identifiers: ({len(result)}, page {page}): {result}")
            else:
                print(f"Model identifiers: ({len(result)}): {result}")
        except Exception as e:
            logger.error(
                f"Error in listing models. {e}{'\n' + e.response.text if isinstance(e, HTTPStatusError) else ''}"
            )
        finally:
            self._cleanup()

    async def _list_tags(self, model_identifier: str):
        return self._model_downloader.list_model_tags(model_identifier=model_identifier)

    async def run_list_tags(self, model_identifier: str):
        try:
            self._initialize()
            result = await self._list_tags(model_identifier=model_identifier)
            print(f"Model tags: ({len(result)}): {result}")
        except Exception as e:
            logger.error(
                f"Error in listing model tags. {e}{'\n' + e.response.text if isinstance(e, HTTPStatusError) else ''}"
            )
        finally:
            self._cleanup()

    async def _model_download(self, model_tag: str):
        self._model_downloader.download_model(model_tag)

    async def run_model_download(self, model_tag: str):
        try:
            self._initialize()
            await self._model_download(model_tag=model_tag)
        except Exception as e:
            logger.error(
                f"Error in downloading model. {e}{'\n' + e.response.text if isinstance(e, HTTPStatusError) else ''}"
            )
        finally:
            self._cleanup()

    async def _hf_list_models(
        self, page: int | None = None, page_size: int | None = None
    ):
        return self._hf_model_downloader.list_available_models(
            page=page, page_size=page_size
        )

    async def run_hf_list_models(
        self, page: int | None = None, page_size: int | None = None
    ):
        try:
            self._initialize()
            result = await self._hf_list_models(page=page, page_size=page_size)
            if page:
                print(f"Model identifiers: ({len(result)}, page {page}): {result}")
            else:
                print(f"Model identifiers: ({len(result)}): {result}")
        except Exception as e:
            logger.error(
                f"Error in listing models. {e}{'\n' + e.response.text if isinstance(e, HTTPStatusError) else ''}"
            )
        finally:
            self._cleanup()

    async def _hf_list_tags(self, model_identifier: str):
        return self._hf_model_downloader.list_model_tags(
            model_identifier=model_identifier
        )

    async def run_hf_list_tags(self, model_identifier: str):
        try:
            self._initialize()
            result = await self._hf_list_tags(model_identifier=model_identifier)
            print(f"Model tags: ({len(result)}): {result}")
        except Exception as e:
            logger.error(
                f"Error in listing model tags. {e}{'\n' + e.response.text if isinstance(e, HTTPStatusError) else ''}"
            )
        finally:
            self._cleanup()

    async def _hf_model_download(self, user_repo_quant: str):
        self._hf_model_downloader.download_model(model_identifier=user_repo_quant)

    async def run_hf_model_download(self, user_repo_quant: str):
        try:
            self._initialize()
            await self._hf_model_download(user_repo_quant=user_repo_quant)
        except Exception as e:
            logger.error(
                f"Error in downloading Hugging Face model. {e}{'\n' + e.response.text if isinstance(e, HTTPStatusError) else ''}"
            )
        finally:
            self._cleanup()


@app.command()
def version():
    """Shows the app version of Ollama downloader."""
    app_handler = OllamaDownloaderCLIApp()
    asyncio.run(app_handler.run_version())


@app.command()
def show_config():
    """Shows the application configuration as JSON."""
    app_handler = OllamaDownloaderCLIApp()
    asyncio.run(app_handler.run_show_config())


@app.command()
def auto_config():
    """Displays an automatically inferred configuration."""
    app_handler = OllamaDownloaderCLIApp()
    asyncio.run(app_handler.run_auto_config())


@app.command()
def list_models(
    page: Annotated[
        Optional[int],
        typer.Option(
            min=1,
            help="The page number to retrieve (1-indexed).",
        ),
    ] = None,
    page_size: Annotated[
        Optional[int],
        typer.Option(
            min=1,
            max=100,
            help="The number of models to retrieve per page.",
        ),
    ] = None,
):
    """Lists all available models in the Ollama library. If pagination options are not provided, all models will be listed."""
    app_handler = OllamaDownloaderCLIApp()
    asyncio.run(app_handler.run_list_models(page=page, page_size=page_size))


@app.command()
def list_tags(
    model_identifier: Annotated[
        str,
        typer.Argument(help="The name of the model to list tags for, e.g., llama3.1."),
    ],
):
    """Lists all tags for a specific model."""
    app_handler = OllamaDownloaderCLIApp()
    asyncio.run(app_handler.run_list_tags(model_identifier=model_identifier))


@app.command()
def model_download(
    model_tag: Annotated[
        str,
        typer.Argument(
            help="The name of the model and a specific to download, specified as <model>:<tag>, e.g., llama3.1:8b. If no tag is specified, 'latest' will be assumed.",
        ),
    ],
):
    """Downloads a specific Ollama model with the given tag."""
    app_handler = OllamaDownloaderCLIApp()
    asyncio.run(app_handler.run_model_download(model_tag=model_tag))


@app.command()
def hf_list_models(
    page: Annotated[
        Optional[int],
        typer.Option(
            min=1,
            help="The page number to retrieve (1-indexed).",
        ),
    ] = 1,
    page_size: Annotated[
        Optional[int],
        typer.Option(
            min=1,
            max=100,
            help="The number of models to retrieve per page.",
        ),
    ] = 25,
):
    """Lists available models from Hugging Face that can be downloaded into Ollama."""
    app_handler = OllamaDownloaderCLIApp()
    asyncio.run(app_handler.run_hf_list_models(page, page_size))


@app.command()
def hf_list_tags(
    model_identifier: Annotated[
        str,
        typer.Argument(
            help="The name of the model to list tags for, e.g., bartowski/Llama-3.2-1B-Instruct-GGUF."
        ),
    ],
):
    """
    Lists all available quantisations as tags for a Hugging Face model that can be downloaded into Ollama.
    Note that these are NOT the same as Hugging Face model tags.
    """
    app_handler = OllamaDownloaderCLIApp()
    asyncio.run(app_handler.run_hf_list_tags(model_identifier=model_identifier))


@app.command()
def hf_model_download(
    user_repo_quant: Annotated[
        str,
        typer.Argument(
            help="The name of the specific Hugging Face model to download, specified as <username>/<repository>:<quantisation>, e.g., bartowski/Llama-3.2-1B-Instruct-GGUF:Q4_K_M.",
        ),
    ],
):
    """Downloads a specified Hugging Face model."""
    app_handler = OllamaDownloaderCLIApp()
    asyncio.run(app_handler.run_hf_model_download(user_repo_quant=user_repo_quant))


def main():
    """Main entry point for the CLI application."""
    # Run the Typer app
    app()


if __name__ == "__main__":  # pragma: no cover
    main()



import logging
import os
from environs import Env
from marshmallow.validate import OneOf
from rich.logging import RichHandler

try:
    from icecream import ic

    ic.configureOutput(includeContext=True)
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa

env = Env()
if os.path.exists(".env"):  # This check is only necessary for Nuitka-compiled binaries.
    env.read_env()  # Read .env file, if it exists


class EnvVar:
    LOG_LEVEL = env.str(
        "LOG_LEVEL",
        default="info",
        validate=OneOf(["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
    ).upper()

    OD_UA_NAME_VER = env.str("OD_UA_NAME_VER", default="ollama-downloader/0.1.1")

    OD_SETTINGS_FILE = env.str(
        "OD_SETTINGS_FILE", default=os.path.join("conf", "settings.json")
    )


logging.basicConfig(
    level=EnvVar.LOG_LEVEL,
    format="%(message)s",
    datefmt="[%X]",
    handlers=[
        RichHandler(
            rich_tracebacks=False, markup=True, show_path=False, show_time=False
        )
    ],
)



import logging
import os
from pathlib import Path
from pydantic import AfterValidator, BaseModel, Field, HttpUrl
from typing import Annotated, ClassVar, List, Optional, Tuple


from ollama_downloader import EnvVar

logger = logging.getLogger(__name__)


class CustomValidators:
    @staticmethod
    def validate_path_as_dir(path_str: str) -> str:
        p = Path(os.path.expanduser(path_str))
        if not p.exists():
            raise ValueError(f"Path '{path_str}' does not exist.")
        if not p.is_dir():
            raise ValueError(f"Path '{path_str}' is not a valid directory.")
        return path_str

    @staticmethod
    def validate_url(url_str: str) -> str:
        parsed_url = HttpUrl(url_str)
        return str(parsed_url)


class OllamaServer(BaseModel):
    url: Annotated[str, AfterValidator(CustomValidators.validate_url)] = Field(
        default="http://localhost:11434/",
        description="URL of the Ollama server.",
    )
    api_key: Optional[str] = Field(
        default=None,
        description="API key for the Ollama server, if required.",
    )
    remove_downloaded_on_error: bool = Field(
        default=True,
        description="Whether to remove downloaded files if the downloaded model cannot be found on the Ollama server, or the Ollama server cannot be accessed.",
    )


class OllamaLibrary(BaseModel):
    models_path: Annotated[
        str, AfterValidator(CustomValidators.validate_path_as_dir)
    ] = Field(
        # Windows environment variables: https://learn.microsoft.com/en-us/windows/deployment/usmt/usmt-recognized-environment-variables
        default=os.path.join(
            "~",
            ".ollama",
            "models",
        ),
        description="Path to the Ollama models on the filesystem. This should be a directory where model BLOBs and manifest metadata are stored.",
    )
    registry_base_url: Annotated[str, AfterValidator(CustomValidators.validate_url)] = (
        Field(
            default="https://registry.ollama.ai/v2/library/",
            description="URL of the remote registry for Ollama models.",
        )
    )
    library_base_url: Annotated[str, AfterValidator(CustomValidators.validate_url)] = (
        Field(
            default="https://ollama.com/library/",
            description="Base URL for the Ollama library. This is used to web scrape model metadata.",
        )
    )
    verify_ssl: Optional[bool] = Field(
        default=True,
        description="Whether to verify SSL certificates when connecting to the Ollama server or registry. Set to False to disable SSL verification (not recommended for production use).",
    )
    timeout: Optional[float] = Field(
        default=120.0,
        description="Timeout for HTTP requests to the Ollama server or registry, in seconds.",
    )
    user_group: Optional[Tuple[str, str]] = Field(
        default=None,
        description="A tuple specifying the username and the group that should own the Ollama models path. If not provided, the current user and group will be used.",
    )


class AppSettings(BaseModel):
    ollama_server: OllamaServer = Field(
        default=OllamaServer(),
        description="Settings for the Ollama server connection.",
    )
    ollama_library: OllamaLibrary = Field(
        default=OllamaLibrary(),
        description="Settings for accessing the Ollama library and storing locally.",
    )

    _instance: ClassVar = None

    def __new__(cls: type["AppSettings"]) -> "AppSettings":
        if cls._instance is None:
            # Create instance using super().__new__ to bypass any recursion
            instance = super().__new__(cls)
            cls._instance = instance
        return cls._instance

    @staticmethod
    def load_or_create_default(
        settings_file: str = EnvVar.OD_SETTINGS_FILE,
    ) -> "AppSettings | None":
        """
        Load settings from the configuration file, or create default settings if the file does not exist.

        Returns:
            AppSettings: The application settings loaded from the configuration file,
            or default settings if the file does not exist.
        """
        settings = AppSettings.load_settings(settings_file)
        if settings is None:
            # This will be a singleton instance
            settings = AppSettings()
            if not AppSettings.save_settings(settings, settings_file):
                return None
        return settings

    @staticmethod
    def load_settings(
        settings_file: str = EnvVar.OD_SETTINGS_FILE,
    ) -> "AppSettings | None":
        """
        Load settings from the configuration file.

        Returns:
            AppSettings: The application settings loaded from the configuration file.
            If the file does not exist or cannot be parsed, returns None.
        """
        try:
            with open(settings_file, "r") as f:
                # Parse the JSON file into the AppSettings model
                return_value = AppSettings.model_validate_json(f.read())
            return return_value
        except FileNotFoundError:
            logger.error(f"Configuration file {settings_file} not found.")
        except Exception as e:
            logger.error(f"Error loading settings from {settings_file}. {e}")
        return None

    @staticmethod
    def save_settings(
        settings: "AppSettings",
        settings_file: str = EnvVar.OD_SETTINGS_FILE,
    ) -> bool:
        """
        Save the application settings to the configuration file.

        Returns:
            bool: True if settings were saved successfully, False otherwise.
        """
        try:
            config_dir = os.path.dirname(settings_file)
            if config_dir and not os.path.exists(config_dir):
                os.makedirs(config_dir, exist_ok=False)
            with open(settings_file, "w") as f:
                f.write(settings.model_dump_json(indent=4))
            logger.info(f"Settings saved to {settings_file}")
            return True
        except Exception as e:
            logger.error(f"Error saving settings to {settings_file}. {e}")
            return False


class ImageManifestConfig(BaseModel):
    mediaType: str = Field(
        ...,
        description="The media type of the image manifest configuration.",
    )
    size: int = Field(
        ...,
        description="The size of the image manifest configuration in bytes.",
    )
    digest: str = Field(
        ...,
        description="The digest of the image manifest configuration, used for content addressing.",
    )


class ImageManifestLayerEntry(BaseModel):
    mediaType: str = Field(
        ...,
        description="The media type of the layer.",
    )
    size: int = Field(
        ...,
        description="The size of the layer in bytes.",
    )
    digest: str = Field(
        ...,
        description="The digest of the layer, used for content addressing.",
    )
    # Note that these URLs may not be present in all manifests and may not be possible to validate as HttpUrls.
    urls: Optional[list[str]] = Field(
        default=None,
        description="Optional list of URLs where the layer can be downloaded from. This is useful for layers that are hosted on multiple locations.",
    )


class ImageManifest(BaseModel):
    # See: https://distribution.github.io/distribution/spec/manifest-v2-2/#image-manifest
    schemaVersion: int = Field(
        ...,
        description="The schema version of the image manifest.",
    )
    mediaType: str = Field(
        ...,
        description="The media type of the image manifest.",
    )
    config: ImageManifestConfig = Field(
        ...,
        description="Configuration for the image manifest, including media type, size, and digest.",
    )
    layers: Optional[List[ImageManifestLayerEntry]] = Field(
        None,
        description="List of layers in the image manifest, each with its media type, size, and digest.",
    )
